/*
 *   MAPA - basic tools for manipulation with a line
 *
 *   (c) 2000 Jan Lana (jlan3028@ss1000.ms.mff.cuni.cz)
 *
 *   $Id: line.c,v 1.25 2001/04/11 12:01:42 mapa Exp $
 */

#include <unistd.h>
#include <string.h>

#include "ml.h"
#include "mlconfig.h"


MLLine *ml_line_new (MLLineAtom *atom, MLTLine type)
{
  MLLine *line;
  
  g_assert (atom != NULL);
  g_assert (atom->line == NULL);
  
  line = g_new (MLLine,1);
  line->len = atom->len;
  line->tip[0] = atom->tip[0];
  line->tip[1] = atom->tip[1];
  line->type = type;

  line->cycle = 0;
  line->penalty = atom->penalty;
  line->width = atom->width;
  
  line->avg_black = atom->len;
  line->avg_space = 0;
  line->pieces = 0;

  line->connect[0] = NULL;
  line->connect[1] = NULL;

  line->atoms = g_ptr_array_new ();
  g_ptr_array_add (line->atoms, atom);
  
  atom->line = line;
  
  return line;
}



void ml_line_free (MLLine *line)
{
  guint i;

  if (line) {
    for (i = 0; i < line->atoms->len; i++)
      ((MLLineAtom*)g_ptr_array_index (line->atoms, i))->line = NULL;
    
    g_ptr_array_free (line->atoms, TRUE);  
    g_free (line);
  }
}


MLFFloat ml_line_direct (const MLLine *line, guint8 tip)
{
  const MLLineAtom *atom;
  MLFFloat direct;
  MLCoord line_tip;
  guint i;
  
  line_tip = line->tip[tip];
  
  i = tip ? line->atoms->len - 1 : 0;
  

  for (;;) {
    atom = (const MLLineAtom*)g_ptr_array_index (line->atoms, i);
    
    direct = ml_lineAtom_direct (atom, (line_tip == atom->tip[1]) ? 1 : 0);
    
    if (direct != ML_DIRECTF_UNKNOWN) 
      return direct;
    
    line_tip = line_tip == atom->tip[1] ? atom->tip[0] : atom->tip[1];
    
    if (tip) {
      if (i-- == 0)
	return ML_DIRECTF_UNKNOWN;
    } else {
      if (++i == line->atoms->len)
	return ML_DIRECTF_UNKNOWN;
    }
  }
}


void ml_line_append (MLLine *line, guint8 tip, MLLine *line2, guint8 tip2,
    MLLineAtom *glue)
{
  guint len, i;
  guint items, sum_black, sum_space;
  MLLineAtom *atom;
#ifdef G_HAVE_GINT64
  guint64 width;
#else
  double width;
#endif
  
  g_assert (glue->line == NULL);
  g_assert (line->tip[tip] == glue->tip[0] || line->tip[tip] == glue->tip[1]);
  g_assert (line2->tip[tip2]==glue->tip[0] || line2->tip[tip2]==glue->tip[1]);
  
  if (line == line2) {
    /* smycka */
    line->len += glue->len;
    line->tip[1] = line->tip[0];
    line->cycle = 1;
      
    g_ptr_array_add (line->atoms, glue);
    
    glue->line = line;
    
    return;
  }

  /* spojeni */
  len = line->atoms->len;
  g_ptr_array_set_size (line->atoms, len + 1 + line2->atoms->len);

  if (tip) {
    g_ptr_array_index (line->atoms, len) = glue;
  } else {
    g_memmove (line->atoms->pdata + 1 + line2->atoms->len, line->atoms->pdata, 
	sizeof (gpointer) * len);

    g_ptr_array_index (line->atoms, line2->atoms->len) = glue;
  }
  
  if (tip == 1 && tip2 == 0)
    for (i = 0; i < line2->atoms->len; i++) 
      g_ptr_array_index (line->atoms, i + len + 1) = 
	g_ptr_array_index (line2->atoms, i);
  else if (tip == 1 && tip2 == 1)
    for (i = 0; i < line2->atoms->len; i++) 
      g_ptr_array_index (line->atoms, i + len + 1) = 
	g_ptr_array_index (line2->atoms, line2->atoms->len - i - 1);
  else if (tip == 0 && tip2 == 0)
    for (i = 0; i < line2->atoms->len; i++) 
      g_ptr_array_index (line->atoms, line2->atoms->len - i - 1) = 
	g_ptr_array_index (line2->atoms, i);
  else {
    g_assert (tip == 0 && tip2 == 1);
  
    for (i = 0; i < line2->atoms->len; i++) 
      g_ptr_array_index (line->atoms, line2->atoms->len - i - 1) = 
	g_ptr_array_index (line2->atoms, line2->atoms->len - i - 1);
  }

  /* update statistik */
  line->len += glue->len + line2->len;
  line->tip[tip] = line2->tip[tip2 ? 0 : 1];
  line->penalty += glue->penalty + line2->penalty;

  /* ... tloustka */
#ifdef G_HAVE_GINT64
  width = G_GINT64_CONSTANT(0);
#else
  width = 0.0;
#endif
  len = 0;
  
  for (i = 0; i < line->atoms->len; i++) {
    atom = (MLLineAtom*)g_ptr_array_index (line->atoms, i);

    if (atom->width > 0) {
#ifdef G_HAVE_GINT64
      width += (guint64)(atom->width * (guint64)(atom->len + 1));
#else
      width += 1.0 * (atom->width * (atom->len + 1));
#endif
      
      len += atom->len + 1;
    }
  }
  
  line->width = width / len;
  
  /* ... avg_* */
  items = 0;
  sum_black = sum_space = 0;
  
  for (i = 0; i < line->atoms->len; i++) {
    atom = (MLLineAtom*)g_ptr_array_index (line->atoms, i);
   
    if (atom->width > 0 || atom->type.glue == MLTG_CROSS) 
      sum_black += atom->len;
    else if (atom->type.glue != MLTG_OVERPRINT) {
      sum_space += atom->len;
      items++;
    }
  }
 
  line->avg_black = sum_black / (items+1);
  line->avg_space = items ? sum_space / items : 0;
  line->pieces = items;
  
  /* update pripojeneho */
  glue->line = line;
  for (i = 0; i < line2->atoms->len; i++)
    ((MLLineAtom*)g_ptr_array_index (line2->atoms, i))->line = line;
}
